/* 
*   clca: Color Linear Cellular Automata
*   87-91  John H. DuBois III
*   88/02/21 last DOS rev
*   See README for complete description.
*/

#ifdef __MSDOS__
#include <dos.h>
#include <conio.h>
#else
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/console.h>
#define ctrlbrk(func) signal(SIGINT,func)
#endif
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include "clca.h"
char *UsageMess[] = {
"CLCA         by    John H. DuBois III",
"Usage: clca [-a<argfile>] [-b<background>] [-g<generations>] [-i<init vals>]",
"            [-v<vidmode>] [-m<mutation period>] [-p<palette>] [-r<rules>]",
"            [-l<line nap>] [-n<screen nap>] [-s] [-w<width>]",
NULL};
char *Comment = "@(#) clca 1.1 96/07/03";

void CleanExit(int foo)
{
    if (RestoreMode()) {
        puts("Error in RestoreMode.");
        exit(1);
    }
    exit(0);
}

main(int argc,char *argv[])
{
    int Background = DEF_BACK;	/* Color of background, set to default */
    int Palette = DEF_PAL;	/* Palette set to default */
    int Width = MAX_CELLS;	/* Set number of cells to max possible */
    /* Number of generations to print, set to default value
     which indicates that printing should continue until program is halted */
    int Gens = 0;
    int MutPeriod = DEF_MUT;	/* Mutation period, set to default */
    int Scroll = 0;		/* Whether screen should be scrolled */
    int Rules[10];	/* Rules used for determining next value of cells */
    int Cells[MAX_CELLS];	/* Array of integers representing cells */
    long Time;			/* Used in seeding random number generator */
    ParamList Params;		/* Params from arg file */
    EnvListType EnvList;	/* List of environments that specify an lca */
    unsigned char PackedCells[MAX_CELLS >> 2];
    int RestoreMode();
    int VidRows, VidCols;
    int VidMode = CGA;     /* Set default vid mode to highest res possible */
				/* currently only handles cga */
    int LineNap = 0,ScreenNap = 0;  /* nap between each line & screen, in ms */

    EnvList.Next = NULL;                /* Initially no environments given */
    Params.Next = NULL;                 /* Initally no params have been read */
    srand((unsigned) time(&Time));    /* Seed random number generator */
    RandInit(Rules,10,3);               /* Initially make rules random */
    RandInit(Cells,MAX_CELLS,3);         /* Start with random cells */
    /* Use flags to set up variables used to generate lca */
    if (!InterpretFlags(argv + 1,argc - 1,&Background,&Palette,&Width,Rules,
     &Gens,&MutPeriod,&Scroll,&VidMode,&LineNap,&ScreenNap,Cells,&Params)) {
        /* If bad values given on command line, print usage */
        PrintStrings(UsageMess);
        exit(2);
    }
    if (Params.Next != NULL && !TranslateParams(&Params,&EnvList)) {
        puts("Bad value given in arg file.");
        PrintStrings(UsageMess);
        exit(2);
    }
    if (SaveMode(0)) {                   /* Save video and tty mode */
        puts("Error in SaveMode.");
        exit(1);
    }
    ctrlbrk(CleanExit);         /* Make ^C restore mode before exiting */

    Gocbreak();

    if (SetVidMode(&VidMode,&VidCols,&VidRows)) { /* Set color graphics mode */
        RestoreMode();
        puts("Error in SetVidMode.");
        exit(1);
     }
    if (Width > VidCols)
	Width = VidCols;
    if (SetPalette(Palette,Background)) {  /* Set palette to selected colors */
        RestoreMode();
        puts("Error in SetPalette.");
        exit(1);
    }
    MakePacked(Cells,PackedCells,Width);
    /* Run lca using command line params or, if given, file params */
    if (EnvList.Next != NULL)
      DriveLCA(&EnvList);
    else
        LCA(Width >> 2,PackedCells,Rules,Gens,MutPeriod,Scroll,LineNap,
	ScreenNap);
    /* When specified num of gens done or keyboard hit, restore mode */
    CleanExit(0);
}

int RandInit(
/* Initialize int array with random values */
int Ints[],       /* Array to be initialized */
int Num,                     /* Size of array */
int Max)                     /* Max value to init to */
{
    int I;
    int ModVal = Max + 1;

    for (I = 0; I < Num; I++)
        Ints[I] = rand() % ModVal;
}

void PrintStrings(char *Strings[])
/* Prints an array of strings, one per line, until a terminating NULL pointer
    is found */
{
    int I;

    for (I = 0; Strings[I] != NULL; I++)
        puts(Strings[I]);
}

void MakePacked(int *Cells,unsigned char PackedCells[],int NumCells)
/* Converts the cell values in Cells (one cell per element of the array)
    to a packed configuration (4 cells per byte) (PackedCells).
    There are NumCells cells and the cells are packed from right to left. */
{
    int I,J,Bytes;
    unsigned char Result;

     Bytes = NumCells >> 2;
     for (I = 0; I < Bytes; I++) {
         Result = 0;
         for (J = 6; J >= 0; J -= 2)
              Result |= ((unsigned char) *(Cells++)) << J;
         PackedCells[I] = Result;
     }
}

void DriveLCA(EnvListType *EnvList)
/* Calls LCA with environments from EnvList */
/* Env List is a circular list of LCA environments from the arg file */
{
    int RulesCopy[10];

    while (!CharPending()) {
         /* Set selected palette */
        SetPalette(EnvList -> Palette,EnvList -> Background);
        /* Make a copy of rules since they may be changed */
        memcpy(RulesCopy,EnvList -> Rules,20);
        LCA(EnvList -> Bytes,EnvList -> PackedCells,RulesCopy,
         EnvList -> Gens,EnvList -> MutPeriod,EnvList -> Scroll,
	 EnvList -> LineNap,EnvList->ScreenNap);
        EnvList = EnvList -> Next;
    }
}

