#ifdef __MSDOS__
#include <conio.h>
#include <ctype.h>
#include <process.h>
#else
#include <sys/types.h>
#include <sys/vtkd.h>
#define getch() getchar()
#endif
#include <string.h>
#include <stdio.h>
#include "clca.h"

void InterpretKey(int *DoMut,int Rules[],int *Scroll,int *MutTimer,
int MutPeriod,int *Line);
void ExpandRules(int *Rules,unsigned char *ExpRules,int NumCalc);
int AllSame(unsigned char Cells[],int Bytes);
void NextGen(unsigned char *PackedCells,unsigned char *ExpRules,int Bytes);
int RestoreMode(void);

void LCA(
/* Generates and prints generations of linear cellular automata */
int Bytes,        /* Number of bytes of cells */
unsigned char PackedCells[],          /* Cell array */
int Rules[],      /* Rules for generating succeeding generations of cells */
int Gens,         /* Number of generations to produce */
int MutPeriod,        /* Mutation period */
int Scroll,
int LineNap,int ScreenNap	/* Nap between lines and screens, in ms */
)       /* Nonzero if screen scrolling should be done */
{
    /* If Gens is 0, the program should continue until halted */
    int NoLimit = !Gens;
    int MutTimer = MutPeriod;         /* Countdown for mutation */
    unsigned char *Middle; /* Pointer to the middle of the cells array */
    int DoMut = MutPeriod;        /* Whether mutation is enabled */
    unsigned char ExpRules[256];          /* Expanded rules */
    int ScrollMode = 0;       /* Whether currently scrolling */
    int Line = 0;         /* Current line being written to */

    if (!DoMut)       /* If mutation was not initially enabled, set default
                             mutation to 200, in case it is enabled */
        MutPeriod = 200;
    Middle = PackedCells + (Bytes >> 1);      /* Find middle of cells array */
    /* Expand rules for 3 summed cells to rules for 4 concatenated cells */
    ExpandRules(Rules,ExpRules,2);
    while (NoLimit || Gens--) {
     /* If no limit, do until program interrupted; otherwise do for num of
         gens specified */
	/* if the first screen has been plotted and scrolling is enabled,
	scroll screen up one line */
	/* also check scroll since it may have been turned off by hitting s */
        if (Scroll & ScrollMode) {
            ScrollUp(Bytes,Line);
            DrawCells(PackedCells,Bytes,199);
	    if (CharPending())  /* If keystroke, carry out key command */
	      InterpretKey(&DoMut,Rules,&Scroll,&MutTimer,MutPeriod,&Line);
        }
        /* If not in scrollmode, write to appropriate column of display */
        else
            DrawCells(PackedCells,Bytes,Line);
	if (Scroll)
	    BufferCells(PackedCells,Bytes,Line);
	if (++Line >= 200) {
	    Line = 0;
	    if (ScreenNap)
		nap((long) ScreenNap);
	    if (CharPending())  /* If keystroke, carry out key command */
	      InterpretKey(&DoMut,Rules,&Scroll,&MutTimer,MutPeriod,&Line);
	    if (Scroll)
	/* If first screen has been displayed and scroll mode should begin,
	    set ScrollMode */
		 ScrollMode = 1;
	}
	if (LineNap)
	    nap((long) LineNap);
        NextGen(PackedCells,ExpRules,Bytes);  /* Find next gen of cells */
        if (DoMut && !(--MutTimer)) { /* If time to do mutation */
            MutTimer = MutPeriod;         /* Reset mutation countdown */
            if (DoMut) {
                RandInit(Rules,10,3);         /* Get new random rules */
                ExpandRules(Rules,ExpRules,2);        /* Expand them */
                /* If all cells are identical, change the middle cell */
                if (AllSame(PackedCells,Bytes))
                    *Middle ^= 0x30;
            }
	    if (DoMut == -1)	/* One-time forced mutation */
		DoMut = 0;
        }
    }
}

void InterpretKey(int *DoMut,int Rules[],int *Scroll,int *MutTimer,
int MutPeriod,int *Line)
/* Carries out key command */
{
    int I,
    x,y;          /* x and y cursor pos */
    int Mode;

    while (CharPending())        /* Carry out key command */
	switch (toupper(getch())) {
	case 'E' :    /* Echo (print current rules on screen) */
#ifdef __TURBOC__
	    CursorLoc(&x,&y);         /* Get cursor pos */
	    GotoXY(30,y);   /* Go to current row, 30th col (3/4 across screen*/
#else
	    RestoreVid();
#endif
	    for (I = 0; I < 10; I++)          /* print rules */
		putchar(Rules[I] + '0');
	    getch();          /* Pause until a key is hit */
#ifndef __TURBOC__
	    SetGraphicsMode();
#endif
	    break;
	case 'F' :
	    *DoMut = !*DoMut;         /* Freeze (disable mutation) */
	    break;
	case 'S' :
	    *Scroll = !*Scroll;       /* Toggle scrolling */
	    if (!*Scroll) {       /* If scrolling disabled, reset
			     mutation and plotting to start at top */
		*Line = 0;
		*MutTimer = MutPeriod;
	    }
	    break;
	case 'M' :
	    *MutTimer = 1;        /* Do immediate mutate */
	    if (!*DoMut)
		*DoMut = -1;
	    break;
	case ' ' :	/* pause */
	    getch();
	    break;
	case 'Q':
	case 3:
	case 0177:
	    RestoreMode();        /* Exit */
	    exit(0);
	    break;
	default:
	    Tone(1000,100L);
	}
}

long Tone(int Freq,long Period)
{
    long Ret;

    ioctl(0,KIOCSOUND,Freq);
    Ret = nap(Period);
    ioctl(0,KIOCSOUND,0);
    return Ret;
}

void ExpandRules(int *Rules,unsigned char *ExpRules,int NumCalc)
/* Expands rules for finding the new value of a cell based on the sum of three
    cells (the cell and its neighbors) (Rules) to rules for finding the values
    of NumCalc cells based on the concatenation of the cells and the cells to
    the right and left of it (ExpRules).  The rules generated assume the cells
    are in order (i.e., LLMMRR, where LL are the bits of the left cell, RR are
    the bits of the right cell, and MM (repeated NumCalc times) are the bits of
    the middle cells).
    To generate the expanded rules, the summed-cells rules are first expanded to
    64 rules for the concatenation of three cells.  These rules are then copied
    to ExpRules in multiple cell positions, with NumCalc cell positions copied
    to in each element of ExpRules.
    To generate the rules for concatenation of three cells, the indexes to the
    summed-cells and cated-cells arrays are initialized to 0.  In each iteration
    a rules is copied from the summed-cells array to the cated-cells array using
    the indices.  Each time there is an overflow in the index to the cated-cells
    array from one cell position (occupying 2 bits) to another, the sum of the
    cells will have been reduced by 2 (the next cell will have been increased
     by 1 while the cell the overflow occured in will have been reduced by 3)
    rather than increasing by 1, so after the summed-cells index is incremented
    3 will be subtracted from it.  If overflow occurs in two cells, the summed-
    cells index is reduced by 6. */
{
#define MAX_CAT (8)       /* Maximum num of concatenated cells that rules can
                                    be generated for */

    unsigned char CatRules[64],       /* Rules for 3 cated cells */
    *NewRule;
    int NumRules,I,J,K,NumRepeat;

    /* Initially increase Rules by 6 because it will be decreased by 6 before
        first being used */
    Rules += 6;
    for (I = 0; I < 64; I++) {        /* Generate rules for 3 cated cells */
        if (!(I & 0x3)) {            /* If overflow from rightmost cell part of
                                                      index, decrease by 3 */
            Rules -= 3;
            if (!(I & 0xf))   /* If overflow from middle cell part of index,
                                                decrease by 3 again */
                Rules -= 3;
        }
        CatRules[I] = *(Rules++);                    /* Copy rule */
    }
    for (I = 0; I < 64; I++)                     /* Copy rules to ExpRules */
        ExpRules[I] = CatRules[I];
    NumRepeat = 4;        /* Initially copy rules over 4 time */
    NumRules = 64;        /* Initially there are 64 rules to copy */
    for (I = 1; I < NumCalc; I++) {       /* Expand rules */
        NewRule = ExpRules + NumRules;  /* Find starting position to duplicate
                                                                 rules */
        /* Copy all rules in ExpRules over to itself 3 times, so that 4
            successive copies of them will exist in ExpRules */
        for (J = 0; J < 3; J++)
            for (K = 0; K < NumRules; K++)
                *(NewRule++) = ExpRules[K];
        NewRule = ExpRules;       /* Start with first rule */
        /* Shift each rule in CatRules left to the next cell position and OR it
            into NumRepeat successive elements of ExpRules */
        for (J = 0; J < 64; J++) {
            CatRules[J] <<= 2;
            for (K = 0; K < NumRepeat; K++)
                *(NewRule++) |= CatRules[J];
        }
        NumRepeat <<= 2;          /* Repeat 4 times as many times next time */
        NumRules <<= 2;       /* Now there are 4 times as many rules */
    }
}

int AllSame(unsigned char Cells[],int Bytes)
/* Returns nonzero if all of the bytes of cells in Cells are identical.
    Bytes is the number of bytes of cells. */
{
    int I;
    unsigned char RightMost = Cells[Bytes - 1];

    for (I = Bytes - 2; I; I--)
        if (Cells[I] != RightMost)
            return 0;
    return(*Cells == RightMost);
}

void NextGen(unsigned char *PackedCells,unsigned char *ExpRules,int Bytes)
/* Generates the next generation of cells in PackedCells using ExpRules */
{
     /* Index can be accessed as an integer using Word, or its hi and lo bytes
        can be accessed individually using Byte.Lo and Byte.Hi */
    union {
        int Word;
        struct {
            unsigned char Lo,Hi;
        } Byte;
    } Index;
    int I;
    /* Holds the high 2 cells (4 bits) of the next generation of cells while
        the low 2 cells are being found */
    unsigned char Result,
    /* The leftmost cell of the cell array, saved to find the rightmost cell */
    LeftCell;

    LeftCell = *PackedCells & 0xc0;       /* Save the leftmost cell */
    /* Start by putting the rightmost cell in the high byte of Index and the
        leftmost byte of cells in the low byte of index */
    Index.Byte.Hi = PackedCells[Bytes - 1] & 0x3;
    Index.Byte.Lo = *PackedCells;
    /* Shift Index left so that the high byte of Index will consist of the
        rightmost cell and the upper 3 cells (6 bits) of the leftmost byte of
        cells */
    Index.Word <<= 6;
    /* Get the next generation of the 2 left cells (they are in the middle 4
        bits of the high byte of Index, so their next gen is gotten when the
        high byte of Index is used as an index to ExpRules) */
    Result = ExpRules[Index.Byte.Hi] << 4;
    /* Shift Index left so that the leftmost byte of cells occupies the hi byte
        of Index.  Index and Result are now in the same state they are each
        time the loop repeats. */
    Index.Word <<= 2;
    for (I = 1; I < Bytes; I++) {
        /* Put the next byte of cells in the lo byte of Index */
        Index.Byte.Lo = PackedCells[I];
        /* Shift Index left so that the hi cell of the byte just put in the lo
            byte of Index will be in the lo 2 bits of the hi byte of Index */
        Index.Word <<= 2;
        /* The middle 4 bits of the hi byte of Index holds the low 2 cells of
           the cell byte being found; the upper 2 bits holds the cell to the
           left of them and the lower 2 bits holds the cell to the right of
           them (from the next byte of cells).  Thus when it is used as an
           index to ExpRules it yields the next generation of the lo 2 cell of
           the byte of cells being found.  This is ORed with the 2 hi cells and
           stored. */
        PackedCells[I - 1] = Result | ExpRules[Index.Byte.Hi];
        /* Shift the middle two cells of the next byte to be found into the hi
            byte of Index */
        Index.Word <<= 4;
        /* Find the next gen for the hi 2 cells and save them in the hi 4 bits
            of Result */
        Result = ExpRules[Index.Byte.Hi] << 4;
        Index.Word <<= 2;         /* Shift the last two bits into the hi byte
                                     of Index */
    }
    /* The rightmost byte of cells uses the leftmost cell as its neighbor to
        the right */
    Index.Byte.Lo = LeftCell;
    Index.Word <<= 2;   /* Shift that cell into the lo 2 bits of the hi byte */
    /* Find and store the rightmost byte of cells */
    PackedCells[Bytes - 1] = Result | ExpRules[Index.Byte.Hi];
}
