#ifdef __MSDOS__
#include <dos.h>
#include <conio.h>
#endif
#include "clca.h"
#include <stdio.h>
#include <string.h>

#define VIDEO 0x10		  /* Video interrupt num */

int RestoreMode(void)
/* Restores video mode as saved by SaveSetMode; returns 0 (for use if called
	by ^C interrupt handler) */
{
	SaveSetMode(1);
	return(0);
}

SaveSetMode(int Func)
/* If Func = 0, saves current video mode in a static var.  Otherwise, sets
	(restores) video mode to value saved in the var. */
{
	static int SaveMode;

	if (Func)
		SetVidMode(SaveMode);
	else
		SaveMode = GetVidMode();
}

int GetVidMode(void)
/* Returns video mode currently in effect */
{
	union REGS regs;

	regs.h.ah = 15;					 /* Get video mode */
	int86(VIDEO,&regs,&regs);		  /* Call ROM-BIOS video service */
	return((int) regs.h.al);		  /* Return video mode */
}

void SetVidMode(int Mode)
/* Set video mode to Mode */
/* For CGA:
		  0 = BW40		  1 = 16 col 40		  2 = BW80		  3 = 16 col 80
		  4 = 4 col 320		  5 = 4 grey 320		  6 = bw 640		  */
{
	union REGS regs;

	regs.h.ah = 0;					 /* Set video mode */
	regs.h.al = (unsigned char) Mode;
	int86(VIDEO,&regs,&regs);		  /* Call ROM-BIOS video service */
}

void SetPalette(int Palette,int Background)
/* Sets medium - res palette to Palette and background color to Background */
/* Palette 0: 0 = background color; 1 = Green; 2 = Red; 3 = Brown.
	Palette 1: 0 = background color; 1 = Cyan; 2 = Magenta; 3 = White.
	Background: 4 bits; format IRGB (intesity, red, green, blue) */
{
	union REGS regs;

	regs.h.ah = 11;					 /* Set palette */
	regs.h.bh = 0;					 /* Set background (set palette subservice) */
	regs.h.bl = (unsigned char) Background;		  /* Background color */
	int86(VIDEO,&regs,&regs);		  /* Call ROM-BIOS video service */
	regs.h.ah = 11;					 /* Set palette */
	regs.h.bh = 1;					 /* Set palette (set palette subservice) */
	regs.h.bl = (unsigned char) Palette;					 /* Palette ID */
	int86(VIDEO,&regs,&regs);		  /* Call ROM-BIOS video service */
}

void GotoXY(unsigned x,unsigned y)
/* Move cursor to position x,y */
{
union REGS regs;

	if (x > 79 || y > 24) {
		printf("Cursor locating error: x = %d; y = %d\n",x,y);
		getch();
	}
	else {
		regs.h.ah = 2;		  /* Set cursor position */
		regs.h.dh = y;
		regs.h.dl = x;
		regs.h.bh = 0;		  /* Video page 0 */
		int86(VIDEO,&regs,&regs);		  /* Call ROM-BIOS video service */
	}
}

void CursorLoc(int *x,int *y)
/* Get current cursor location */
{
union REGS regs;

	regs.h.ah = 3;					 /* read cursor position */
	regs.h.bh = 0;					 /* Video page 0 */
	int86(VIDEO,&regs,&regs);		  /* Call ROM-BIOS video service */
	*x = regs.h.dl;
	*y = regs.h.dh;
}

void ScrollUp()
/* Scrolls the graphics screen up one line. */
{
	int Offset;
	int Bytes = 0; /* faked out */

	for (Offset = 0; Offset < 8000; Offset += 80) {
		movedata(0xba00,Offset,0xb800,Offset,Bytes);
		movedata(0xb805,Offset,0xba00,Offset,Bytes);
	}
}
